#ifndef __UPF_ABI_H__
#define __UPF_ABI_H__

#include <stdlib.h>
#include <string.h>

#include <upf/platform.h>
#include <upf/version.h>
#include <upf/context.h>

#if defined(__UPF_UNIX__)
    /* FIXME:
     * use something like HAVE_STDINT_H (but independent of config.h!) */
    #include <stdint.h>
    typedef uint8_t        upf_uint8_t;
    typedef uint16_t       upf_uint16_t;
    typedef uint32_t       upf_uint32_t;
#elif defined(__UPF_WIN32__)
    typedef unsigned char  upf_uint8_t;
    typedef unsigned short upf_uint16_t;
    typedef unsigned       upf_uint32_t;
#endif

/* ----------------------------------------------------------------------
                              ABI data structures
   ---------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_UPF_PRAGMA_PACK_GNU)
#pragma pack(1)
#elif defined(_UPF_PRAGMA_PACK_WIN32)
#pragma pack(push,1)
#else
#error "Must specify packed structs pragma for your compiler!"
#endif

struct _upf_Object;


/* This should be defined to some reasonable value because it is allocated
   on the *stack* every time a method is called: */
#define _UPF_ARGUMENTS_CHUNK_SIZE    512

typedef struct _upf_ArgumentsChunk
{
    upf_uint8_t                     storage[_UPF_ARGUMENTS_CHUNK_SIZE];
    struct _upf_ArgumentsChunk *next;
} upf_ArgumentsChunk;

typedef struct
{
    size_t              size;
    unsigned            chunkPos;
    unsigned            totalPos; /* used only for safety checks */
    upf_ArgumentsChunk *currChunk;
    upf_ArgumentsChunk  data;
} upf_Arguments;

typedef struct
{
    upf_uint32_t data1, data2, data3, data4;
    upf_uint16_t version_major;
    upf_uint16_t version_minor;
} upf_IID;

typedef upf_uint32_t upf_CID;

#define UPF_CHECK_IID(requested, available) \
         ((requested).data1 == (available).data1 && \
          (requested).data2 == (available).data2 && \
          (requested).data3 == (available).data3 && \
          (requested).data4 == (available).data4 && \
          (requested).version_major == (available).version_major && \
          (requested).version_minor <= (available).version_minor)
          

typedef unsigned int upf_MethodID;
typedef unsigned int upf_ContextID;

typedef void (_UPF_CALLING_CONVENTION *upf_DispatchFunction)(void*,
                                                             upf_MethodID,
                                                             upf_Arguments*);

typedef struct
{
    void*                data;
    upf_DispatchFunction function;
} upf_InterfaceData;

/* NB: queryInterface function returns its return value via 'retval' parameter
       on purpose! We can't return structures in binary compatible way (Borland
       is known to use different mechanism from MSVC and Mingw even when __cdecl
       or __stdcall is specified). */
typedef void (_UPF_CALLING_CONVENTION *upf_QueryInterfaceFunction)(
                                                             struct _upf_Object *object,
                                                             const upf_IID* iface,
                                                             upf_InterfaceData *retval);

typedef struct _upf_Object
{
    upf_QueryInterfaceFunction  queryInterface;
    void                       *data;
    upf_ContextID               nativeContext;
    unsigned long               refCnt;
} upf_Object;


typedef struct
{
    unsigned long ABI_version;
    upf_Object*   manager;
    upf_Object*   serviceManager;

    struct
    {
        upf_Object*    (_UPF_CALLING_CONVENTION *create_instance)(upf_CID classID);
        void           (_UPF_CALLING_CONVENTION *marchaller_init)(upf_Arguments *args);
        void           (_UPF_CALLING_CONVENTION *marchaller_seek0)(upf_Arguments *args);
        void           (_UPF_CALLING_CONVENTION *marchaller_destroy)(upf_Arguments *args);
        void           (_UPF_CALLING_CONVENTION *marchaller_write)(upf_Arguments *args, void *data, size_t len);
        void           (_UPF_CALLING_CONVENTION *marchaller_read)(upf_Arguments *args, void *data, size_t len);
        upf_ContextID  (_UPF_CALLING_CONVENTION *get_context)(const char *context_string);
        upf_CID        (_UPF_CALLING_CONVENTION *alloc_new_cid)(void);
        upf_CID        (_UPF_CALLING_CONVENTION *classname_to_cid)(const char *classname);
        int            (_UPF_CALLING_CONVENTION *merge_typeinfo)(const upf_uint8_t *typeinfo);
    } funcs;
} upf_middlewareLayer;

#if defined(_UPF_PRAGMA_PACK_GNU)
#pragma pack()
#elif defined(_UPF_PRAGMA_PACK_WIN32)
#pragma pack(pop)
#else
#error "Must specify packed structs pragma for your compiler!"
#endif


#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
    #define _UPF_EXTERN_C   extern "C"
    #define _UPF_EXTERN     extern "C"
#else
    #define _UPF_EXTERN_C
    #define _UPF_EXTERN     extern
#endif




/* ----------------------------------------------------------------------
                              UPF public API
   ---------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

#ifdef UPF_BUILDING_MODULE
    _UPF_EXTERN upf_middlewareLayer *upf_middleware;
    /* NB: we can't use same symbol name for upf_context because 
           ELF linker would merge all the symbols from .so plugins
           into the symbol defined in libupf.so
     */
    _UPF_EXTERN upf_ContextID __upf_module_context;
    #define upf_context __upf_module_context
#else
    _UPF_EXPORTED_SYMBOL upf_middlewareLayer *upf_middleware;
    #define upf_init() upf_init_(UPF_RUNTIME_CONTEXT)
    _UPF_EXPORTED_SYMBOL int _UPF_CALLING_CONVENTION upf_init_(const char *app_context);
    _UPF_EXPORTED_SYMBOL int _UPF_CALLING_CONVENTION upf_done(void);
    _UPF_EXPORTED_SYMBOL upf_ContextID upf_context;

    #ifdef UPF_BUILDING_CORE
        /* NB: for the same reason, we want to use distinct upf_context
               for libupf itself. */
        _UPF_EXTERN upf_ContextID __upf_internal_context;
        #define upf_context __upf_internal_context
    #endif
#endif


#ifdef __cplusplus
}
#endif



/* ----------------------------------------------------------------------
                          DLL modules loading
   ---------------------------------------------------------------------- */

#define _UPF_DLL_ACTION_INIT_CLASSES   0
#define _UPF_DLL_ACTION_LOAD           1
#define _UPF_DLL_ACTION_UNLOAD         2

#define _UPF_DLL_ENTRY_POINT           __upf_DllEntryPoint
#define _UPF_DLL_ENTRY_POINT_NAME      _UPF_STRINGIZE(_UPF_DLL_ENTRY_POINT)
typedef int (_UPF_CALLING_CONVENTION * _upf_dllEntryPointFunc)(
	         upf_middlewareLayer *middleware,
		 int                  action
	    );


#endif /* __UPF_ABI_H__ */
