#ifndef __UPF_CXX_H__
#define __UPF_CXX_H__

/** \file cxx.h
    C++ UPF headers.
    \note Don't include this file directly, use \c <upf/upf.h> instead.
  */

#include <stdio.h>
#include <upf/abi.h>
#include <string>
#include <vector>

namespace upf
{
typedef upf_InterfaceData InterfacePtr;
typedef upf_IID IID;
}

inline bool operator==(const ::upf::IID& id1, const ::upf::IID& id2)
{
    return id1.data1 == id2.data1 && id1.data2 == id2.data2 &&
           id1.data3 == id2.data3 && id1.data4 == id2.data4 &&
           id1.version_major == id2.version_major &&
           id1.version_minor == id2.version_minor;
}

inline bool operator!=(const ::upf::IID& id1, const ::upf::IID& id2)
{
    return !(id1 == id2);
}


namespace upf {

class IObject;
class IManager;
class IWriteableClassInfo;

template<class T> class Ptr; // fwd reference

// ----------------------------------------------------------------------------
// Helpers:
// ----------------------------------------------------------------------------

template<class T> struct InterfaceTraits;


enum MultiPurposeHookMode
{
   _UPF_HOOK_MODE_INIT_CLASSINFO      = 1,
   _UPF_HOOK_MODE_QUERY_INTERFACE     = 2
};

// This is a dirty hack neccessary because some compilers (MSVC, as usual)
// can't do partial template specialization
struct IWriteableClassInfoWrapper
{
    virtual void addInterface(const IID& iid) = 0;
};
template<class T>
struct IWriteableClassInfoWrapperImpl : public IWriteableClassInfoWrapper
{
    IWriteableClassInfoWrapperImpl(T *t) : m_t(t) {}
    void addInterface(const IID& iid) { m_t->addInterface(iid); }
    T *m_t;
};


struct MultiPurposeHookArgs
{
    MultiPurposeHookMode mode;
    union
    {
        struct
        {
            IWriteableClassInfo *ci;
            IWriteableClassInfoWrapper *ciWrapper;
        } ici;
        struct
        {
            const IID*   iid;
            InterfacePtr retval;
        } qi;
    } u;
};



class CxxBase
{
    public:
        virtual upf_Object* _upf_getABIObject() = 0;
};

template<class T> class Proxy;


inline bool checkIID(const IID& requested, const IID& available)
{
    return UPF_CHECK_IID(requested, available);
}

template<class T> inline T* _makeCxxObject(upf_Object *object)
{
    if (object == NULL)
        return NULL;
    upf_InterfaceData i;
    object->queryInterface(object, &InterfaceTraits<T>::iid(), &i);
    if (object->nativeContext == upf_context)
    {
        return static_cast<T*>(i.data);
    }
    else
    {
        if (i.data == NULL)
            return NULL;
        else
            return new Proxy<T>(object, i);
    }
}

// NB: this is yet another dirty hack. We don't need to make this function
//     a template because we will only use it for T=upf_Object. But it can't
//     be regular inlined function because calling convention can't be specified
//     for inline functions in win32 compilers. The function can't be non-inline,
//     either, because gcc would require implementation file then and we want to
//     avoid having to link against C++ library at any cost because of binary
//     incompability.
template<typename T> 
void _UPF_CALLING_CONVENTION _CxxABIQueryInterface(T *object, const upf_IID *iid,
                                                   upf_InterfaceData *retval)
{
    *retval = (static_cast<IObject*>(object->data))->queryInterface(*iid);
}



/** \internal
    Queries the object for given interface. 
    \note Use upf::Ptr instead of directly calling this function.
    \return pointer to interface implementation or NULL if the object does not
            provide this interface.
 */
template<class T> inline T* queryInterface(IObject *obj)
{
    if (obj == NULL)
        return NULL;
    else
    {
        InterfacePtr i = obj->queryInterface(InterfaceTraits<T>::iid());
        upf_Object *abi = obj->_upf_getABIObject();
        if (abi->nativeContext == upf_context)
        {
            return reinterpret_cast<T*>(i.data);
        }
        else
        {
            if (i.data == NULL)
                return NULL;
            else
                return new Proxy<T>(abi, i);
        }
    }
}



// ----------------------------------------------------------------------------
// upf implementation-side API:
// ----------------------------------------------------------------------------

#define _UPF_DECLARE_CLASS_NO_FRIENDS(classname) \
    private:\
        typedef classname _upfClassType; \
        upf_Object _upf_object; \
        static void _upfMultiPurposeHook(::upf::MultiPurposeHookArgs *data, classname *instance); \
        static void _upf_initClass(::upf::IWriteableClassInfo *classinfo); \
        static ::upf::CID _upf_getCID(); \
        void _upf_initInstance(); \
    public: \
        upf_Object* _upf_getABIObject() { return &_upf_object; } \
        void incRef(); \
        void decRef(); \
        ::upf::InterfacePtr queryInterface(const ::upf::IID& iface); \
        ::upf::CID getClassID();

// Visual C++ 6 is either too old to implement ISO C++ or it ignores it:
#ifdef _MSC_VER
    #define _UPF_DECLARE_CLASS_FRIENDS_DECL(classname) \
        friend class ::upf::StdFactory< classname >; \
        friend class ::upf::ClassRegistrator< classname >;
#else
    #define _UPF_DECLARE_CLASS_FRIENDS_DECL(classname) \
        typedef ::upf::StdFactory< classname > _upfFactory; \
        typedef ::upf::ClassRegistrator< classname > _upfRegistrator; \
        friend class _upfFactory; \
        friend class _upfRegistrator;
#endif

/** Declares that a class is available to UPF system. This macro will
    add some internally used variables and implementations of IObject
    methods. This macro must be part of class declaration.

    Example:
    \code
    namespace foo { 
      class Bar : public my::IBar  {
        ...
        UPF_DECLARE_CLASS(Bar)
      };
    }
    \endcode
    
    \see \ref cxx_tutorial, UPF_IMPLEMENT_CLASS
    \hideinitializer
 */
#define UPF_DECLARE_CLASS(classname) \
    _UPF_DECLARE_CLASS_FRIENDS_DECL(classname) \
    _UPF_DECLARE_CLASS_NO_FRIENDS(classname)


#define _UPF_DO_IMPLEMENT_CLASS(classname, prefix) \
    prefix void classname::incRef() \
    { \
        _upf_object.refCnt++; \
    } \
    prefix void classname::decRef() \
    { \
        if (--_upf_object.refCnt == 0) \
        { \
            delete this; \
        } \
    } \
    prefix ::upf::InterfacePtr classname::queryInterface(const ::upf::IID& iface) \
    { \
        ::upf::MultiPurposeHookArgs data; \
        data.mode = ::upf::_UPF_HOOK_MODE_QUERY_INTERFACE; \
        data.u.qi.iid = &iface; \
        data.u.qi.retval.data = NULL; \
        _upfMultiPurposeHook(&data, this); \
        return data.u.qi.retval; \
    } \
    prefix void classname::_upf_initInstance() \
    { \
        _upf_object.refCnt = 0; \
        _upf_object.data = static_cast< ::upf::IObject* >(this); \
        _upf_object.nativeContext = upf_context; \
        _upf_object.queryInterface = ::upf::_CxxABIQueryInterface; \
    } \
    prefix void classname::_upf_initClass(::upf::IWriteableClassInfo *classinfo) \
    { \
        if (classinfo == NULL) return; \
        ::upf::MultiPurposeHookArgs data; \
        ::upf::IWriteableClassInfoWrapperImpl< ::upf::IWriteableClassInfo > wrapper(classinfo); \
        data.mode = ::upf::_UPF_HOOK_MODE_INIT_CLASSINFO; \
        data.u.ici.ci = classinfo; \
        data.u.ici.ciWrapper = &wrapper; \
        classinfo->setCID(_upf_getCID()); \
        _upfMultiPurposeHook(&data, NULL); \
    } \
    prefix ::upf::CID classname::_upf_getCID() \
    { \
        static ::upf::CID my_cid = 0; \
        if (my_cid == 0) \
            my_cid = upf_middleware->funcs.alloc_new_cid(); \
        return my_cid; \
    } \
    prefix ::upf::CID classname::getClassID() \
    { \
        return _upf_getCID(); \
    } \
    prefix void classname::_upfMultiPurposeHook(::upf::MultiPurposeHookArgs* data, classname *instance)


/** Declares that the class implements interface \a iface. Can be only used
    inside a UPF_IMPLEMENT_CLASS block.

    \see \ref cxx_tutorial, UPF_IMPLEMENT_CLASS
    \hideinitializer
 */
#define UPF_INTERFACE(iface) \
    if (::upf::InterfaceTraits< iface >::query(instance, data)) return;


/** Assigns property to the class. Can be only used inside
    a UPF_IMPLEMENT_CLASS block.
    
    \see \ref cxx_tutorial, UPF_IMPLEMENT_CLASS, \ref properties
    \hideinitializer
 */
#define UPF_PROPERTY(name, value) \
    if (data->mode == ::upf::_UPF_HOOK_MODE_INIT_CLASSINFO) \
        data->u.ici.ci->addProperty(name, value);


/** Declares that class named \a classname must be available. (I.e. call to
    upf::create("\a classname") would create its instance.) If this condition
    is not met, the class will \em not be registered in UPF classes registry.
    
    \param classname Name of required class in C++ syntax, without quotes.
           An example:
           \code
           UPF_REQUIRES_CLASS(upf::impl::ServiceManager)
           \endcode
  
    \todo NOT YET IMPLEMENTED!
    \see \ref cxx_tutorial, UPF_IMPLEMENT_CLASS
    \hideinitializer
 */
#define UPF_REQUIRES_CLASS(classname)
    // FIXME -- TODO

#define _UPF_NOTHING 

/** This macro expands to implementation of UPF boilerplate code. It must
    be used for every class that contains UPF_DECLARE_CLASS in its declaration.
    UPF_IMPLEMENT_CLASS is followed by { } brackets that contains UPF_INTERFACE
    and UPF_PROPERTY declarations.

    \code
    UPF_IMPLEMENT_CLASS(foo::Bar)
    {
      UPF_INTERFACE(my::IBar)

      UPF_PROPERTY("Description", "A simple calculator implementation")
      UPF_PROPERTY("Author",      "Joe Hacker")

      UPF_REQUIRES_CLASS(foo::Helper)
    }
    \endcode        

    \see \ref cxx_tutorial,
         UPF_DECLARE_CLASS, UPF_PROPERTY, UPF_INTERFACE, UPF_REQUIRES_CLASS
    \hideinitializer
 */
#define UPF_IMPLEMENT_CLASS(classname)  _UPF_DO_IMPLEMENT_CLASS(classname, _UPF_NOTHING)





/** Register class \a classname in UPF system. You must either call this macro
    or use runtime-loaded DLL and UPF_DLL_MODULE macro, otherwise the class
    won't be visible in UPF!
    Example:
    \code
    upf::init();
    UPF_REGISTER_CLASS(MyClass);
    ...
    \endcode

    \see UPF_DLL_MODULE, UPF_EXPORTED_CLASS
    \hideinitializer
 */
#define UPF_REGISTER_CLASS(classname) \
    ::upf::ClassRegistrator< classname >::registerClass( #classname )

#ifndef UPF_BUILDING_MODULE
/** Declarator used to enumerate classes exported from DLL. The macro expands to
    code that declares and implements DLL's entry point. An example:
    \code
    UPF_DLL_MODULE()
    {
        UPF_EXPORTED_CLASS(ClassOne)
        UPF_EXPORTED_CLASS(ClassTwo)
    }
    \endcode
    
    \note \c UPF_BUILDING_MODULE macro must be defined for this macro to be
          available.  

    \note This macro may be used only once per DLL/shared library. If your
          module DLL consists of several source files, only one of them can
          contain UPF_DLL_MODULE block (and it may occur only once in the file,
          too).
    
    \see \ref cxx_tutorial, UPF_EXPORTED_CLASS, UPF_ON_LOAD, UPF_ON_UNLOAD
    \hideinitializer
 */
#define UPF_DLL_MODULE() \
   const char * _UPF_ERROR = You_must_define_UPF_BUILDING_MODULE_macro_on_your_command_line;
#else
#define UPF_DLL_MODULE() \
    upf_middlewareLayer *upf_middleware; \
    upf_ContextID __upf_module_context; \
    static void __upf_DllEntryPointImpl(int action, int *retval); \
    _UPF_ALWAYS_EXPORTED_SYMBOL int _UPF_CALLING_CONVENTION \
                       _UPF_DLL_ENTRY_POINT(upf_middlewareLayer *middleware, \
				            int action) \
    { \
        if (!UPF_CHECK_ABI_VERSION(middleware->ABI_version)) \
            return 0; \
        if (action == _UPF_DLL_ACTION_LOAD) \
	{ \
            upf_middleware = middleware; \
            __upf_module_context = \
                middleware->funcs.get_context(UPF_RUNTIME_CONTEXT); \
        } \
	int retval = 1; \
        __upf_DllEntryPointImpl(action, &retval); \
	return retval; \
    } \
    static void __upf_DllEntryPointImpl(int action, int *retval)


#endif

/** Declare that class \a classname should be exported from DLL. For use only
    inside UPF_DLL_MODULE statement. 
    
    \note \c UPF_BUILDING_MODULE macro must be defined for this macro to be
          available.  
    \see \ref cxx_tutorial, UPF_REGISTER_CLASS
    \hideinitializer
 */
#define UPF_EXPORTED_CLASS(classname) \
    if (action == _UPF_DLL_ACTION_INIT_CLASSES) \
        { UPF_REGISTER_CLASS(classname); }


// This trait class makes it possible to use both functions returning
// bool and void as onLoad/onUnload functions:
template<typename T> inline bool _initFunctionCaller(T (*func)());
template<> inline bool _initFunctionCaller<bool>(bool (*func)())
    { return func(); }
template<> inline bool _initFunctionCaller<void>(void (*func)())
    { func(); return true; }


/** Function \a func will be called immediately after the DLL module is
    loaded. For use only inside UPF_DLL_MODULE statement. \a func may
    either return bool value (the module won't be loaded if any of
    UPF_ON_LOAD functions return <tt>false</tt>) or not return any value.
    
    \note \c UPF_BUILDING_MODULE macro must be defined for this macro to be
          available.  
    \note This function is called \em before classes from this module
          are registered in UPF registry!
    \see \ref cxx_tutorial, UPF_REGISTER_CLASS, UPF_ON_UNLOAD
    \hideinitializer
 */
#define UPF_ON_LOAD(func) \
    if (action == _UPF_DLL_ACTION_LOAD && \
        !::upf::_initFunctionCaller(func)) { *retval = 0; return; }
   
/** Function \a func will be called before the DLL module is unloaded.
    For use only inside UPF_DLL_MODULE statement. \a func may
    either return bool value (the module won't be loaded if any of
    UPF_ON_LOAD functions return <tt>false</tt>) or not return any value.

    \note \c UPF_BUILDING_MODULE macro must be defined for this macro to be
          available.  
    \see \ref cxx_tutorial, UPF_REGISTER_CLASS, UPF_ON_LOAD
    \hideinitializer
 */
#define UPF_ON_UNLOAD(func) \
    if (action == _UPF_DLL_ACTION_UNLOAD && \
        !::upf::_initFunctionCaller(func)) { *retval = 0; return; }

}; // end of namespace 'upf'


#endif
