#ifndef __UPF_CXX_PTR_H__
#define __UPF_CXX_PTR_H__

#include <assert.h>

namespace upf
{

class IObject;

// -----------------------------------------------------------------------------------
// Smart pointer:
// -----------------------------------------------------------------------------------


/** UPF's smart pointer. \c upf::Ptr<T> holds pointer of type \c T* and overrides
    \c operator-> so that it can be used as if it were \c T*.
    upf::Ptr's function is two-fold:
    - It manages reference counting for UPF objects by calling IObject::incRef when
      new pointer value is assigned to the variable and IObject::decRef when old
      value is replaced by a new one or from destructor. You don't have to explicitly
      delete objects, they are deleted automatically when their reference count reaches
      zero. In fact you must \b not delete any UPF-managed object yourself!

    - It uses upf::queryInterface to convert between various interfaces implemented by
      the object and to determine if some interface is implemented at all. Interface
      queries are transparent to the user: if you assign a pointer to Ptr<T> variable,
      it will either contain \c T* pointer pointing to object's valid interface or
      \c NULL if the object does not implement interface \c T.
      
    \param T interface type. This \e must be interface, not implementation class type!

    A simple example is here:
    \code
    {
        upf::Ptr<IFoo> foo = upf::create("MyFooImpl");
        upf::Ptr<upf::IClassInfo> x = foo; 
        // x == NULL now because IFoo does not implement upf::IClassInfo
        {
            upf::Ptr<upf::IObject> base = foo;  // foo's ref count increased by 1 (to 2)
            // 'base' contains valid pointer, every interface implements IObject
        } // 'base' ptr destroyed, ref cnt decreased to 1
    } // 'foo' ptr destroyed, ref cnt decreased to 0, the object itself is deleted
    \endcode

    \see \ref ref_cnt
 */
template<class T> class Ptr
{
public:
    /// Default ctor.
    Ptr() : m_obj(NULL) {}

    /// Copy ctor
    Ptr(const Ptr& ptr)
    {
        m_obj = ptr.m_obj;
        if (m_obj) m_obj->incRef();
    }

#if !defined(_MSC_VER) || _MSC_VER > 1200
    /** "Copy" ctor from non-<tt>T</tt> smart pointer.
        upf::queryInterface is called to determine
        if interface \c T is supported by \a ptr. If it is not, the pointer is assigned
        \c NULL value instead of \a ptr.
     */
    template<class O> Ptr(const Ptr<O>& ptr)
    {
        m_obj = upf::queryInterface<T>((O*)ptr);
        if (m_obj) m_obj->incRef();
    }
#endif

    /** Ctor from non-<tt>T</tt> ordinary pointer.
        upf::queryInterface is called to determine
        if interface \c T is supported by \a obj. If it is not, the pointer is assigned
        \c NULL value instead of \a obj.
     */
    Ptr(IObject *obj)
    {
        m_obj = upf::queryInterface<T>(obj);
        if (m_obj) m_obj->incRef();
    }

    /// Ctor (doesn't call upf::queryInterface)
    Ptr(T *obj)
    {
        m_obj = obj;
        if (m_obj) m_obj->incRef();
    }

    /// Destructor
    ~Ptr()
    {
        if (m_obj) m_obj->decRef();
    }

    /// \c operator-> allows us to use Ptr as ordinary pointers    
    T* operator->() const
    {
        assert(m_obj != NULL);
        return m_obj;
    }

    /** \c operator= for non-<tt>T</tt> types. upf::queryInterface is called to determine
        if interface \c T is supported by \a obj. If it is not, the pointer is assigned
        \c NULL value instead of \a obj.
     */
    const Ptr& operator=(IObject *obj)
    { 
        if (m_obj == obj)
            return *this;
        T* oldobj = m_obj;
        m_obj = upf::queryInterface<T>(obj);
        if (m_obj) m_obj->incRef();
        if (oldobj) oldobj->decRef();
        return *this;
    }

    /// Fast version of operator= that does not call upf::queryInterface
    const Ptr& operator=(const Ptr& ptr)
    {
        return *this = ptr.m_obj;
    }

#if !defined(_MSC_VER) || _MSC_VER > 1200
    /** \c operator= for non-<tt>T</tt> types. upf::queryInterface is called to determine
        if interface \c T is supported by the object hold in \a ptr. If it is not, 
        the pointer is assigned \c NULL value instead of \a ptr.
     */
    template <class O> const Ptr& operator=(const Ptr<O>& ptr)
    {
        const Ptr& x = *this = (O*)ptr;
        return x;
    }
#endif

    /// Fast version of operator= that does not call upf::queryInterface
    const Ptr& operator=(T *obj)
    { 
        if (m_obj == obj)
            return *this;
        if (m_obj) m_obj->decRef();
        m_obj = obj;
        if (m_obj) m_obj->incRef();
        return *this;
    }

    /// Cast to \c T*
    operator T*() const { return m_obj; }

private:
    T *m_obj;
};


// NB: Ptr<IObject> needs a version without IObject operators: 
template<> class Ptr<IObject>
{
public:
    Ptr() : m_obj(NULL) {}

    Ptr(const Ptr& ptr)
    {
        m_obj = ptr.m_obj;
        if (m_obj) m_obj->incRef();
    }

#if !defined(_MSC_VER) || _MSC_VER > 1200
    template<class O> Ptr(const Ptr<O>& ptr)
    {
        m_obj = (O*)ptr;
        if (m_obj) m_obj->incRef();
    }
#endif

    Ptr(IObject *obj)
    {
        m_obj = obj;
        if (m_obj) m_obj->incRef();
    }

    ~Ptr()
    {
        if (m_obj) m_obj->decRef();
    }

    IObject* operator->() const
    {
        assert(m_obj != NULL);
        return m_obj;
    }

    const Ptr& operator=(IObject *obj)
    { 
        if (m_obj == obj)
            return *this;
        if (m_obj) m_obj->decRef();
        m_obj = obj;
        if (m_obj) m_obj->incRef();
        return *this;
    }

    const Ptr& operator=(const Ptr& ptr)
    { 
        if (m_obj == ptr.m_obj)
            return *this;
        if (m_obj) m_obj->decRef();
        m_obj = ptr.m_obj;
        if (m_obj) m_obj->incRef();
        return *this;
    }

#if !defined(_MSC_VER) || _MSC_VER > 1200
    template <class O> const Ptr& operator=(const Ptr<O>& ptr)
    {
        return *this = (O*)ptr;
    }
#endif

    operator IObject*() const
        { return m_obj; }

private:
    IObject *m_obj;
};


}

#endif
