
#
# UPF XML interface descriptions backend to omniidl. Based on
# idldump.py example backend from OmniORB, original copyright notice
# follows:
#

#                           Package   : omniidl
# idldump.py                Created on: 1999/10/29
#			    Author    : Duncan Grisby (dpg1)
#
#    Copyright (C) 1999 AT&T Laboratories Cambridge
#
#  This file is part of omniidl.
#
#  omniidl is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
#  02111-1307, USA.
#
# Description:
#   
#   Back-end which just dumps the IDL tree


from omniidl import idlast, idltype, idlutil, idlvisitor, output
import sys, string, os, tempfile, re, md5
import inchelper


# ----------------- Helper utils --------------------------------

def xmlScopedName(node):
    return string.join(node.scopedName(), '.')

def xmlEncodeString(s):
    return s.replace('&','&amp;').replace('"', '&quot;').\
             replace('<', '&lt;').replace('>', '&gt;')



def collectPragmas(node):

    def collList(list):
        c = []
        for n in list:
            c += collectPragmas(n)
        return c
    
    def fixPragmas(node, pragmas):
        return map(lambda x: (xmlScopedName(node), x), pragmas)

    if isinstance(node, idlast.AST):
        return map(lambda x: ('',x), node.pragmas()) + collList(node.declarations())
    elif isinstance(node, idlast.Module) and node.mainFile():
        return fixPragmas(node, node.pragmas()) + collList(node.definitions())
    elif isinstance(node, idlast.Interface) and node.mainFile():
        return fixPragmas(node, node.pragmas())
    else:
        return []



# ----------------- Output class used to write data to XML file ------

class XMLMegaFile:
    def __init__(self, name):
        self.lines = []
        self.name = name

    def addLines(self, lines):
        self.lines += lines

    def close(self):
        if self.name == '-':
            f = sys.stdout
        else:
            f = open(self.name, 'wt')
        f.write("""\
<?xml version="1.0"?>
<UPF-IDL xmlns="http://upf.sourceforge.net/UPF-IDL">
""")
        f.writelines(self.lines)
        f.write("""\
</UPF-IDL>
""")
        if self.name == '-':
            f = None
        else:
            f.close()

class BundledOutputStream(output.Stream):
    def __init__(self, xmlfile):
        self.xmlfile = xmlfile
        self.tmpName = tempfile.mktemp('-upf.xml')
        self.file = open(self.tmpName, 'wt')
        output.Stream.__init__(self, self.file, 2)
        self.inc_indent()

    def close(self):
        """MUST call this when done with the file, so that it is
           dumped to the mega-XML file"""
        self.dec_indent()
        self.file.close()
        f = open(self.tmpName, 'rt')
        self.xmlfile.addLines(f.readlines())
        f.close()
        os.remove(self.tmpName)



# ----------------- IDL tree visitor -----------------------------

class UPFVisitor (idlvisitor.AstVisitor, idlvisitor.TypeVisitor):

    def __init__(self, xmlfile, includeDirs):
        self.xmlfile = xmlfile
        self.includeDirs = includeDirs
        self.st = None
        self.oldStreams = []
        
    def openStream(self):
        if self.st != None:
            self.oldStreams.append(self.st)
        self.st = BundledOutputStream(self.xmlfile)
        
    def closeStream(self):
        self.st.close()
        if len(self.oldStreams) > 0:
            self.st = self.oldStreams[-1]
            del self.oldStreams[-1]
        else:
            self.st = None

    def visitAST(self, node):
        self.openStream()
        self.st.out('<filename idl="@idl@" xml="@xml@.upface"/>',
                    xml=os.path.splitext(node.file())[0],
                    idl=node.file())

        for i in inchelper.findIncludes(node):
            header=i
            local='1'
            for d in self.includeDirs:
                if header[:len(d)] == d:
                    header = header[len(d)+1:]
                    local='0'
            if local == '1':
                dir1 = os.path.dirname(node.file())
                if dir1 == '': dir1 = '.'
                dir1 += '/'
                dir2 = os.path.dirname(header)
                if dir2 == '': dir2 = '.'
                dir2 += '/'
                common = os.path.commonprefix([dir1,dir2])
                header = (dir2+os.path.basename(header))[len(common):]
            self.st.out('<include file="@fn@" local="@local@"/>',
                        fn=header, local=local)

        for scope, pragma in collectPragmas(node):
            self.st.out('<pragma scope="@scope@" text="@txt@"/>',
                        scope=scope, txt=xmlEncodeString(pragma.text()))
        self.closeStream()

        for n in node.declarations():
            if not n.mainFile():
                self.openStream()
                self.st.out('<UPF-INCLUDED-IDL>')
                self.closeStream()
            n.accept(self)
            if not n.mainFile():
                self.openStream()
                self.st.out('</UPF-INCLUDED-IDL>')
                self.closeStream()

    def visitModule(self, node):
        for n in node.definitions():
            n.accept(self)

    def visitInterface(self, node):
        self.openStream()
        version = node.repoId().split(':')[-1]
        md5sum = md5.new(xmlScopedName(node)).hexdigest()
        self.st.out('<interface id="@iface@" version="@version@" IID="@md5sum@">',
                    iface=xmlScopedName(node), version=version, md5sum=md5sum)
        self.st.inc_indent()
        if len(node.inherits()) != 0:
            for i in node.inherits():
                self.st.out('<super iface="@super@"/>', super=xmlScopedName(i))

        for n in node.contents():
            n.accept(self)

        self.st.dec_indent()
        self.st.out('</interface>')
        self.closeStream()


    def visitForward(self, node):
        self.openStream()
        self.st.out('<forward_reference interface_id="@id@"/>', id=xmlScopedName(node))
        self.closeStream()


    def visitConst(self, node):
        node.constType().accept(self)
        type = self.__result_type

        if node.constKind() == idltype.tk_enum:
            value = xmlScopedName(node.value())
        else:
            value = str(node.value())
        
        self.openStream()
        self.st.out("""\
<const type="@type@" id="@id@" value="@value@"/>""",
                    type=type, id=xmlScopedName(node), 
                    value=xmlEncodeString(value))
        self.closeStream()


    def visitTypedef(self, node):
        if node.constrType():
            node.aliasType().decl().accept(self)

        node.aliasType().accept(self)
        type  = self.__result_type
        for d in node.declarators():
            self.openStream()
            d.accept(self)
            decl = self.__result_declarator
            self.st.out('<typedef id="@decl@">', decl=decl)
            self.st.inc_indent()
            self.st.out('@type@', type=type)
            self.st.dec_indent()
            self.st.out('</typedef>')
            self.closeStream()


    def visitStruct(self, node):
        self.openStream()
        self.st.out('<struct id="@id@">', id=xmlScopedName(node))
        self.st.inc_indent()

        for m in node.members():
            if m.constrType():
                self.st.inc_indent()
                m.memberType().decl().accept(self)
                self.st.dec_indent()

            m.memberType().accept(self)
            type = self.__result_type
            for d in m.declarators():
                d.accept(self)
                self.st.out('<item type="@type@" id="@id@"/>',
                            id=self.__result_declarator,
                            type=type)

        self.st.dec_indent()
        self.st.out('</struct>')
        self.closeStream()



    def visitException(self, node):
        self.openStream()
        self.st.out('<exception id="@id@">',
                    id = xmlScopedName(node))

        for m in node.members():
            if m.constrType():
                self.st.inc_indent()
                m.memberType().decl().accept(self)
                self.st.dec_indent()

            m.memberType().accept(self)
            type = self.__result_type
            for d in m.declarators():
                d.accept(self)
                self.st.out('<item type="@type@" id="@id@"/>',
                            id=self.__result_declarator,
                            type=type)

        self.st.out('</exception>')
        self.closeStream()



    def visitUnion(self, node):
        self.openStream()
        if node.constrType():
            self.st.out('<union id="@id@" switch="',
                        id=xmlScopedName(node))
            self.st.inc_indent()
            node.switchType().decl().accept(self)
            self.st.out('">')
            self.st.dec_indent()
        else:
            node.switchType().accept(self)
            stype = self.__result_type
            self.st.out('<union id="@id@" switch="@stype@">',
                        id=xmlScopedName(node), stype=stype)

        for c in node.cases():
            if c.constrType():
                self.st.inc_indent()
                c.caseType().decl().accept(self)
                self.st.dec_indent()

            self.st.inc_indent()
            for l in c.labels():
                if l.default():
                    self.st.out('<case default="1">')
                else:
                    if l.labelKind() == idltype.tk_enum:
                        lv = xmlScopedName(l.value())
                    else:
                        lv = str(l.value())
                        
                    self.st.out('<case value="@lv@">', 
                                lv=xmlEncodeString(lv))
                    
            c.caseType().accept(self)
            type = self.__result_type
            c.declarator().accept(self)
            decl = self.__result_declarator
            self.st.inc_indent()
            self.st.out('<item type="@type@" id="@id@"/>',
                        id=decl,
                        type=type)
            self.st.dec_indent()
            self.st.out('</case>')
            self.st.dec_indent()

        self.st.out('</union>')
        self.closeStream()



    def visitEnum(self, node):
        self.openStream()
        self.st.out('<enum id="@id@">', id = xmlScopedName(node))
        self.st.inc_indent()
        for e in node.enumerators(): 
            self.st.out('<value name="@name@"/>', name=xmlScopedName(e))
        self.st.dec_indent()
        self.st.out('</enum>')
        self.closeStream()



    def visitAttribute(self, node):
        if node.readonly(): readonly = ' readonly="1"'
        else: readonly = ''

        node.attrType().accept(self)
        type = self.__result_type

        for d in node.declarators():
            self.st.out('<attribute@readonly@ type="@type@" id="@id@"/>', 
                        id=xmlScopedName(d), readonly=readonly, type=type)



    def visitOperation(self, node):
        if node.oneway(): oneway=' oneway="1"'
        else: oneway=''
        
        node.returnType().accept(self)
        rtype = self.__result_type
        self.st.out('<method id="@id@" returns="@rtype@"@oneway@>',
                    id=xmlScopedName(node), rtype=rtype, oneway=oneway)
        self.st.inc_indent()

        for p in node.parameters():
            p.paramType().accept(self)
            ptype = self.__result_type
            if p.is_in(): is_in='1'
            else: is_in='0'
            if p.is_out(): is_out='1'
            else: is_out='0'
            self.st.out(\
'<param type="@ptype@" name="@name@" in="@is_in@" out="@is_out@"/>',\
                name=p.identifier(), ptype=ptype,\
                is_in=is_in, is_out=is_out)

        for r in node.raises():
            self.st.out('<raises exception="@name@"/>', name=xmlScopedName(r))

        self.st.dec_indent()
        self.st.out('</method>')



    def visitNative(self, node):
        self.openStream()
        self.st.out('<native id="@id@"/>', id=xmlScopedName(node))
        self.closeStream()



    def visitDeclarator(self, node):
        l = [xmlScopedName(node)]
        if len(node.sizes()) > 0:
            print "Error: foo[bar] type declarators not supported yet!"
            sys.exit(1)
        #for s in node.sizes():
        #    l.append("[" + str(s) + "]")

        self.__result_declarator = string.join(l, "")


    ttsMap = {
        idltype.tk_void:       "void",
        idltype.tk_short:      "short",
        idltype.tk_long:       "long",
        idltype.tk_ushort:     "unsigned short",
        idltype.tk_ulong:      "unsigned long",
        idltype.tk_float:      "float",
        idltype.tk_double:     "double",
        idltype.tk_boolean:    "boolean",
        idltype.tk_char:       "char",
        idltype.tk_octet:      "octet",
        idltype.tk_any:        "any",
        idltype.tk_TypeCode:   "CORBA::TypeCode",
        idltype.tk_Principal:  "CORBA::Principal",
        idltype.tk_longlong:   "long long",
        idltype.tk_ulonglong:  "unsigned long long",
        idltype.tk_longdouble: "long double",
        idltype.tk_wchar:      "wchar"
        }

    def visitBaseType(self, type):
        self.__result_type = self.ttsMap[type.kind()]

    def visitStringType(self, type):
        if type.bound() == 0:
            self.__result_type = "string"
        else:
            self.__result_type = \
                '<bounded bound="%s">string</bounded>' % str(type.bound())

    def visitWStringType(self, type):
        if type.bound() == 0:
            self.__result_type = "wstring"
        else:
            self.__result_type = \
                '<bounded limit="%s">wstring</bounded>' % str(type.bound())

    def visitSequenceType(self, type):
        type.seqType().accept(self)
        if type.bound() == 0:
            self.__result_type = "<sequence>" + self.__result_type + "</sequence>"
        else:
            self.__result_type = '<sequence bound="%s">%s</sequence>' %\
                                 (str(type.bound()),self.__result_type)

    def visitFixedType(self, type):
        self.__result_type = '<fixed digits="%s" scal="%s"/>'%\
                             (str(type.digits()),str(type.scale()))

    def visitDeclaredType(self, type):
        self.__result_type = xmlScopedName(type.decl())



def run(tree, args):
    includes = []
    for a in args:
        l = string.split(a,'=')
        key = l[0]
        value = l[1]
        if key == 'output':
            filename = value
        if key == 'include':
            includes.append(value)
    
    xmlfile = XMLMegaFile(filename)
    dv = UPFVisitor(xmlfile, includes)
    tree.accept(dv)
    xmlfile.close()
