
import xml.dom.minidom, string, sys, os.path, typeinf


PRIMITIVE_TYPES = ['void',
                   'short', 'long', 'long long', 
                   'unsigned short', 'unsigned long', 'unsigned long long',
                   'float', 'double', 'long double', 'boolean', 'octet', 'char',
                   'wchar', 'string', 'wstring']


# Entity types:
ENTITY_INTERFACE   = typeinf.ENTITY_INTERFACE
ENTITY_TYPEDEF     = typeinf.ENTITY_TYPEDEF
ENTITY_NATIVE_TYPE = typeinf.ENTITY_NATIVE_TYPE
ENTITY_STRUCT      = typeinf.ENTITY_STRUCT
ENTITY_CONST       = typeinf.ENTITY_CONST
ENTITY_UNION       = typeinf.ENTITY_UNION
ENTITY_ENUM        = typeinf.ENTITY_ENUM

class Entity:
    """Single interface entity (e.g. class, structure). In short, entities are
       chunks of source code that may depend on each other (e.g. class Foo derived
       from Bar requires that Bar code is before Foo code)"""
    def __init__(self, node):
        self.id = node.getAttribute('id') # Unique string identifier (e.g. module.class)
        self.deps = []                    # IDs of entity's dependencies
        self.children = []                # Children entity objects      
        self.ent_type = None

class NativeTypeEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_NATIVE_TYPE

    def serialize(self, allEnts):
        return typeinf.serializeNativeType(self)


class TypedefEntity(Entity):
    SEQUENCE = typeinf.TYPEDEF_SEQUENCE
    SIMPLE   = typeinf.TYPEDEF_SIMPLE

    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_TYPEDEF
        self.isPrimitive = 0
        node.normalize()
        sequence = node.getElementsByTagName('sequence')
        if len(sequence) > 0:
            self.kind = self.SEQUENCE
            self.type = sequence[0].firstChild.data.strip()
        else:
            self.kind = self.SIMPLE
            self.type = node.firstChild.data.strip()
        if self.type not in PRIMITIVE_TYPES:
            self.deps.append(self.type)

    def serialize(self, allEnts):
        return typeinf.serializeTypedef(self)


class ConstEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_CONST
        self.isPrimitive = 1
        node.normalize()
        self.value = node.getAttribute('value')
        self.type = node.getAttribute('type')
        if self.type not in PRIMITIVE_TYPES:
            self.deps.append(self.type)

    def serialize(self, allEnts):
        return typeinf.serializeConst(self)


class InterfaceEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_INTERFACE
        self.super      = []
        self.methods    = []
        self.version    = node.getAttribute('version')
        self.IID        = node.getAttribute('IID')
        self.singleImpl = None

        for super in node.getElementsByTagName('super'):
            s = super.getAttribute('iface')
            self.super.append(s)
            self.deps.append(s)

        for op in node.getElementsByTagName('method'):
            params = []
            for p in op.getElementsByTagName('param'):
                tp = p.getAttribute('type')
                if tp not in PRIMITIVE_TYPES:
                    self.deps.append(tp)
                params.append(typeinf.MethodParameter(tp,
                                                 p.getAttribute('name'),
                                                 p.getAttribute('in'),
                                                 p.getAttribute('out')))
            ret = op.getAttribute('returns')
            if ret not in PRIMITIVE_TYPES:
                self.deps.append(ret)
            self.methods.append(typeinf.Method(op.getAttribute('id'),
                                               ret, params))

    def serialize(self, allEnts):
        return typeinf.serializeInterface(self, allEnts)


class StructEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_STRUCT
        self.variables = []
        for v in node.getElementsByTagName('item'):
            tp = v.getAttribute('type')
            self.variables.append(typeinf.Variable(tp, v.getAttribute('id')))
            if tp not in PRIMITIVE_TYPES:
                self.deps.append(tp)

    def serialize(self, allEnts):
        return typeinf.serializeStruct(self)


class UnionEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_UNION
        self.switchType = node.getAttribute('switch')
        self.cases = []
        for v in node.getElementsByTagName('case'):
            item = v.getElementsByTagName('item')[0]
            tp = item.getAttribute('type');
            self.cases.append(typeinf.UnionCase(v.getAttribute('value'),
                                                tp, item.getAttribute('id')))
            if tp not in PRIMITIVE_TYPES:
                self.deps.append(tp)

    def serialize(self, allEnts):
        return typeinf.serializeUnion(self)


class EnumEntity(Entity):
    def __init__(self, node):
        Entity.__init__(self, node)
        self.ent_type = ENTITY_ENUM
        self.values = []
        for v in node.getElementsByTagName('value'):
            self.values.append(v.getAttribute('name'))

    def serialize(self, allEnts):
        return typeinf.serializeEnum(self)


class Pragma:
     def __init__(self, scope, text):
        self.scope = scope
        self.text = text


class ParsedFile:
    """Structure representing one XML file. Includes all entities, dependencies
       etc., so that it can be directly translated into e.g. C++ code."""
    def __init__(self):
        self.forwardRefs = []    # Forward references to interfaces
        self.entities = {}       # Dictionary of all parsed entities
        self.toplevel = []       # Ordered list of toplevel entities
        self.includes = []       # List of included files
                                 # (tuples: (filename,is_local) )
        self.allEntities = {}    # Dictionary of entities in *all* files, not
                                 # only parsed one but all included as well
        self.idlFile = None      # Filename of the source IDL
        self.pragmas = []        # List of #pragmas found in IDL file
        self.__ordered = []

    def parseFile(self, file):
        try:
            dom = xml.dom.minidom.parse(file)
        except IOError:
            sys.stderr.write('Error: cannot read input file!\n')
            sys.exit(1)
        assert dom.documentElement.tagName == 'UPF-IDL'
        self.__parseXML(dom.documentElement)
        dom.unlink()
        self.__interpretPragmas()
        self.__orderEntities()
    
    def __addEntity(self, ent):
        self.entities[ent.id] = ent
        self.__ordered.append(ent.id)
    
    def __interpretPragmas(self):
        for p in self.pragmas:
            spl = p.text.split()
            # handle "#pragma single_impl iface_name impl_name":
            if len(spl) > 0 and spl[0] == 'single_impl':
                if self.entities.has_key(p.scope):
                    self.entities[p.scope].singleImpl = spl[2]
            

    def __parseXML(self, rootNode):
        for n in rootNode.childNodes:
            if n.nodeType != n.ELEMENT_NODE: continue

            if n.tagName == 'typedef':
                self.__addEntity(TypedefEntity(n))

            elif n.tagName == 'interface':
                self.__addEntity(InterfaceEntity(n))

            elif n.tagName == 'forward_reference':
                # fwd reference is special, it doesn't have unique id
                self.forwardRefs.append(n.getAttribute('interface_id'))

            elif n.tagName == 'native':
                self.__addEntity(NativeTypeEntity(n))

            elif n.tagName == 'struct':
                self.__addEntity(StructEntity(n))

            elif n.tagName == 'union':
                self.__addEntity(UnionEntity(n))

            elif n.tagName == 'enum':
                self.__addEntity(EnumEntity(n))

            elif n.tagName == 'const':
                self.__addEntity(ConstEntity(n))

            elif n.tagName == 'UPF-INCLUDED-IDL':
                inc = ParsedFile()
                inc.__parseXML(n)
                self.allEntities.update(inc.allEntities)

            elif n.tagName == 'include':
                is_local = int(n.getAttribute('local'))
                hdr_name = n.getAttribute('file')
                self.includes.append((hdr_name,is_local))
                    
            elif n.tagName == 'filename':
                self.idlFile = n.getAttribute('idl')

            elif n.tagName == 'pragma':
                self.pragmas.append(Pragma(n.getAttribute('scope'), n.getAttribute('text')))

        self.allEntities.update(self.entities)
        
    def __orderEntities(self):
        """Creates self.toplevel list which contains entity IDs ordered as
           they should appear in generated code and append entities to children
           property of entities (again, sorted)."""

        self.toplevel = []
        for id in self.__ordered:
            parent = string.join(id.split('.')[:-1], '.')
            if parent != '' and self.entities.has_key(parent):
                self.entities[parent].children.append(self.entities[id])
            else:
                self.toplevel.append(self.entities[id])
        self.__ordered = None


def serializeWithDeps(entity, allEntities):
    """Serializes given entity *and all its dependencies* into streams
       with format described in doc/typeinfo-format.txt"""
       
    def findAllEntities(ent, list):
        if not ent in list:
            list.append(ent)
            for d in ent.deps:
                if allEntities.has_key(d):
                    findAllEntities(allEntities[d], list)

    list = []
    findAllEntities(entity, list)
    
    ret = ''
    for e in list:  
        #if e.ent_type == None: continue
        s = e.serialize(allEntities)
        ret += typeinf.encUint32(len(s)) + s
    ret += typeinf.encUint32(0)
    return ret


def serializeAllInDict(ns):
    ret = ''
    for id in ns:
        s = ns[id].serialize(ns)
        ret += typeinf.encUint32(len(s)) + s
    ret += typeinf.encUint32(0)
    return ret
