
import re, os.path
from iface import *
import output

allEntities = None
needsToGeneratePart2 = []


# ---------------------------- C++ helpers ------------------------------------

GENERATED_FOR_ABI_VERSION = '2'

PRIMITIVE_TYPES = {
   'void'               : 'void',
   'short'              : 'short',
   'long'               : 'long',
   'long long'          : 'long long',
   'unsigned short'     : 'unsigned short',
   'unsigned long'      : 'unsigned long',
   'unsigned long long' : 'unsigned long long',
   'float'              : 'float',
   'double'             : 'double',
   'long double'        : 'long double',
   'boolean'            : 'bool',
   'octet'              : 'upf_uint8_t',
   'char'               : 'char',
   'wchar'              : 'wchar_t',
   'string'             : '::std::string',
   'wstring'            : '::std::wstring',
   }

TYPE_INTERFACE = 1
TYPE_PRIMITIVE = 2
TYPE_OTHER     = 3

def cxxGetTypeInfo(tp):
    if tp == 'string' or tp == 'wstring':
        return TYPE_OTHER  # treat std::string as "complicated" objects
    if PRIMITIVE_TYPES.has_key(tp):
    #if not allEntities.has_key(tp):
        return TYPE_PRIMITIVE
    e = allEntities[tp]
    if isinstance(e, InterfaceEntity):
        return TYPE_INTERFACE
    elif isinstance(e, TypedefEntity):
        if e.kind == e.SIMPLE:
            return cxxGetTypeInfo(e.type)
        else:
            return TYPE_OTHER
    else:
        return TYPE_OTHER

def cxxFullId(id, context=''):
    if context == id:
        return id[1+id.rfind('.'):]
    if context != '':
        idL = id.split('.')
        cL = context.split('.')
        if idL[:-1] == cL[:-1]:
            return idL[-1]
    return '::%s' % id.replace('.', '::')

def cxxShortId(id):
    return id.split('.')[-1]

def cxxType(tp, context=''):
    if PRIMITIVE_TYPES.has_key(tp):
        return PRIMITIVE_TYPES[tp]
    else:
        return cxxFullId(tp, context)

def cxxTypeArg(tp, out, context=''):
    info = cxxGetTypeInfo(tp)
    if info == TYPE_PRIMITIVE:
        if out: return '%s&' % cxxType(tp, context)
        else:   return cxxType(tp, context)
    elif info == TYPE_INTERFACE:
        if out: return '::upf::Ptr< %s >&' % cxxFullId(tp, context)
        else:   return '%s*' % cxxFullId(tp, context)
    else:
        if out: return '%s&' % cxxType(tp, context)
        else:   return 'const %s&' % cxxType(tp, context)

def cxxTypeStorage(tp, context=''):
    """For temporary storage of arguments in proxy code"""
    info = cxxGetTypeInfo(tp)
    if info == TYPE_INTERFACE:
        return '::upf::Ptr< %s >' % cxxFullId(tp, context)
    else:
        return cxxType(tp, context)


def cxxTypeRet(tp, context=''):
    # NB: can't return "SomeIface*", must do "upf::Ptr<SomeIface>",
    #     because the object could be deleted otherwise if there wasn't
    #     any reference hold point to it.
    return cxxTypeStorage(tp,context)


def cxxMkMacro(name):
    return re.sub(r'[-./\\]', '_', name.upper())


def cxxMkArray(string):
    out = '\n'
    cnt = 0
    for s in string[:-1]:
        out += '%i,' % ord(s)
        cnt += 1
        if cnt == 20:
            out += '\n'
            cnt = 0
    out += '%i' % ord(string[-1])
    return out


# ------------------------------- Code generator -----------------------------

# Special implementation that is prefixed/postfixed to standard proxy code:
PROXY_SPECIAL_CODE = {
    'upf.IObject.incRef'         : '_proxy_incRef();',
    'upf.IObject.decRef'         : '_proxy_decRef();',
    'upf.IObject.queryInterface' : 'return _proxy_queryInterface(interface_id);',
  }

def genParamsList(op, context):
    params = ''
    if len(op.params) > 0:
        for p in op.params[:-1]:
            params += '%s %s, ' % (cxxTypeArg(p.type, out=p.isOut, context=context), p.name)
        p = op.params[-1]
        params += '%s %s' % (cxxTypeArg(p.type, out=p.isOut, context=context), p.name)
    return params

def generateInterface(st, e):
    if len(e.super) > 0:
        def mkDecl(x):
            if x == 'upf.IObject':
                return 'virtual public %s' % cxxFullId(x)
            else:
                return 'public %s' % cxxFullId(x)
        supers = ' : %s' % string.join(map(mkDecl, e.super), ', ')
    elif e.id == 'upf.IObject':
        supers = ' : public ::upf::CxxBase'
    else:
        supers = ''
    st.out("""\
class @name@@supers@
{
public:
""",   name=cxxShortId(e.id), supers=supers)
    st.inc_indent()

    for child in e.children:
        generateEntity(st, child)
    
    paramsDict = {}
    for op in e.methods:
        st.out('virtual @returns@ @id@(@params@) = 0;',
               returns=cxxTypeRet(op.returns, e.id), 
               id=cxxShortId(op.id), params=genParamsList(op, e.id))

    st.dec_indent()
    st.out('};')

    needsToGeneratePart2.append(e)


def _generateInterfaceProxyTemplate(st, e):
    offset = '0';
    baseclass = 'Base'
    if len(e.super) > 0:
        for super in e.super:
            baseclass = '_ProxyCode_%s< %s, %s >' % \
                        (cxxMkMacro(super), baseclass, offset)
            offset = '%s+__UPF_%s_OPCNT' % (offset, cxxMkMacro(super))
    
    st.out("""
template<class Base, int Offset>
class _ProxyCode_@macro@ : public @baseclass@
{""", baseclass=baseclass, 
      macro=cxxMkMacro(e.id), opCnt = str(len(e.methods)))

    st.inc_indent()
    opNum = 0
    for op in e.methods:
        pre_code = ''
        post_code = ''
        if PROXY_SPECIAL_CODE.has_key(op.id):
            st.out("""\
virtual @returns@ @id@(@params@)
{
    @code@
}""", returns=cxxTypeRet(op.returns), id=cxxShortId(op.id), 
      params=genParamsList(op, ''), code=PROXY_SPECIAL_CODE[op.id])
        else:        
            st.out("""\
virtual @returns@ @id@(@params@)
{
@precode@    ArgumentsOwned _upf_args;""",
                   returns=cxxTypeRet(op.returns), 
                   precode=pre_code,
                   id=cxxShortId(op.id), params=genParamsList(op, ''))
            for p in filter(lambda x: x.isIn, op.params):
                st.out('    _upf_args << %s;' % p.name)
            st.out('    m_dispatch.function(m_dispatch.data, Offset+@opNum@+@offset@, _upf_args.getMarchalled());',
                   opNum=str(opNum), offset=offset)
            for p in filter(lambda x: x.isOut, op.params):
                st.out('    _upf_args >> %s;' % p.name)
            if op.returns != 'void':
                st.out("""\
    @rettype@ _upf_retval;
    _upf_args >> _upf_retval;
    return _upf_retval;""", rettype=cxxTypeRet(op.returns))
            st.out('}')
        opNum += 1

    st.dec_indent()

    st.out("""\
};
#define __UPF_@macro@_OPCNT ((@offset@)+(@opCnt@))""",
             macro=cxxMkMacro(e.id),offset=offset, opCnt=str(len(e.methods)))


def generateInterfacePart2(st, e):
    global allEntities
    st.out('namespace upf {')
    
    for op in e.methods:
        name = '__upf_dispatch__%s' % op.id.replace('.','_') 
        st.out("""\
inline void @name@(@fullname@ *_upf_data, upf_Arguments *_upf_marchalled_args)
{""", name=name, fullname=cxxFullId(e.id))
        if len(op.params) > 0 or op.returns != 'void':
            st.out('    upf::Arguments _upf_args(_upf_marchalled_args);')


        st.inc_indent()
        args=[]
        for p in op.params:
            st.out('@type@ @name@;', name=p.name, type=cxxTypeStorage(p.type))
            args.append(p.name);

        inputParams = filter(lambda p: p.isIn, op.params)
        if len(inputParams) > 0:
            st.out('_upf_args.seek0();')
        for p in inputParams:
            st.out('_upf_args >> @name@;', name=p.name)
        
        argsS = string.join(args, ', ')

        if op.returns == 'void':
            st.out('_upf_data->@opid@(@args@);',
                   ifacename=cxxFullId(e.id), opid=cxxShortId(op.id),args=argsS)
        else:
            st.out('@rettype@ _upf_retval(_upf_data->@opid@(@args@));',
                   ifacename=cxxFullId(e.id), opid=cxxShortId(op.id),
                   rettype=cxxTypeRet(op.returns), args=argsS)

        outputParams = filter(lambda p: p.isOut, op.params)
        if (len(outputParams) > 0 or op.returns != 'void') and len(inputParams) > 0:
            st.out('_upf_args.seek0();')
        for p in outputParams:
            st.out('_upf_args << @name@;', name=p.name)        
        if op.returns != 'void':
            st.out('_upf_args << _upf_retval;')
        if len(outputParams) > 0 or op.returns != 'void':
            st.out('_upf_args.seek0();')

        st.dec_indent()
        st.out('}\n')

    st.out("""
template<> struct InterfaceTraits< @fullname@ >
{
    typedef @fullname@ value_type;

    static void _UPF_CALLING_CONVENTION dispatchFunction(void *obj, upf_MethodID method, upf_Arguments *args)
    {
        @fullname@ *ptr = static_cast< @fullname@* >(obj);""",fullname=cxxFullId(e.id))
    topCnt = '0'
    bottomCnt = '0'
    elseSt = ''
    for super in e.super:
        topCnt = '%s + __UPF_%s_OPCNT' % (topCnt, cxxMkMacro(super))
        st.out("""\
        @elseSt@if (method < (@cnt@))
        {
            InterfaceTraits< @super@ >::dispatchFunction(static_cast< @super@* >(ptr),
                                                         method - (@bottom@), args);
        }""", elseSt=elseSt, cnt=topCnt, bottom=bottomCnt, super=cxxFullId(super))
        elseSt = 'else '
        bottomCnt = topCnt
        
    st.out("""\
        @elseSt@
        {
            static void (*functions[])(@fullname@ *, upf_Arguments *) = {""",
            elseSt=elseSt, fullname=cxxFullId(e.id))
    for op in e.methods:
        st.out('            __upf_dispatch__@id@,', id=op.id.replace('.','_') )
    st.out('            NULL };')
    st.out("""\
            functions[method - (@bottom@)](ptr, args);
        }
    }""", bottom=bottomCnt)
    
    if e.singleImpl != None:
        st.out("""
    static const char* implName()
        { return "@impl@"; }""", impl=e.singleImpl)

    st.out("""
    static const char* name()
        { return "@origID@"; }

    static const IID& iid()
    {
        static IID _iid = {0x@iid1@,0x@iid2@,0x@iid3@,0x@iid4@,@major@,@minor@};
        return _iid;
    }
    
    static const upf_uint8_t *typereg()
    {
        static const upf_uint8_t desc[] = { @description@ };
        return desc;
    }

    static bool query(value_type *instance, MultiPurposeHookArgs *data)
    {
        if (data->mode == _UPF_HOOK_MODE_QUERY_INTERFACE)
        {
            if (upf::checkIID(*(data->u.qi.iid), iid()))
            {
                data->u.qi.retval.data = instance;
                data->u.qi.retval.function = dispatchFunction;
                return true;
            }
        }
        else if (data->mode == _UPF_HOOK_MODE_INIT_CLASSINFO)
        {
            upf_middleware->funcs.merge_typeinfo(typereg());
            data->u.ici.ciWrapper->addInterface(iid());
        }""", 
             origID = e.id,
             major=e.version.split('.')[0], minor=e.version.split('.')[1],
             description=cxxMkArray(serializeWithDeps(e, allEntities)),
             iid1=e.IID[0:8], iid2=e.IID[8:16], iid3=e.IID[16:24], iid4=e.IID[24:32])
    for s in e.super:
        st.out('       if (InterfaceTraits< @super@ >::query(instance, data)) return true;',
               super=cxxFullId(s))
    st.out("""\
        return false;
    }
};""")

    _generateInterfaceProxyTemplate(st, e)

    st.out("""\
template<> class Proxy< @fullname@ > : 
    public _ProxyCode_@macro@< _ProxyBase< @fullname@ >, 0 >
{
public:
    Proxy(upf_Object *object, const upf_InterfaceData& dispatch)
    {
        init_proxy(object, dispatch);
    }
    virtual ~Proxy() {}
};

} // namespace upf""", 
     fullname=cxxFullId(e.id),
     macro=cxxMkMacro(e.id))




def generateTypedef(st, e):
    if e.kind == e.SIMPLE:
        st.out('typedef @type@ @id@;',
               id=cxxShortId(e.id), type=cxxType(e.type, context=e.id))
    elif e.kind == e.SEQUENCE:
        st.out('typedef ::std::vector< @type@ > @id@;',
               id=cxxShortId(e.id), type=cxxTypeStorage(e.type, context=e.id))


def generateConst(st, e):
    type = cxxType(e.type, context=e.id)
    if e.type == 'string':
        st.out('const char @id@[] = "@value@";\n',
               id=cxxShortId(e.id), value=e.value)
    else:
        st.out('const @type@ @id@ = @value@;\n',
               id=cxxShortId(e.id), type=type, value=e.value)


def generateEnum(st, e):
    st.out('enum @id@\n{', id=cxxShortId(e.id))
    st.inc_indent()
    for v in e.values[:-1]:
        st.out('@name@,', name=cxxShortId(v))
    st.out('@name@', name=cxxShortId(e.values[-1]))
    st.dec_indent()
    st.out('};')
    needsToGeneratePart2.append(e)

def generateEnumPart2(st, e):
    st.out("""
namespace upf {
inline Arguments& operator<<(Arguments& holder, const @type@& value)
{
    return holder << (long)value;
}
inline Arguments& operator>>(Arguments& holder, @type@& value)
{
    long x;
    holder >> x;
    value = (@type@)x;
    return holder;
}
}
""", type=cxxFullId(e.id))


def generateStruct(st, e):
    st.out("""\
struct @id@
{
    @id@() {}
""", id=cxxShortId(e.id))
    st.inc_indent()

    def mkArg(v):
        return '%s _%s' % (cxxTypeArg(v.type, 0, e.id), cxxShortId(v.name))
    def mkAssign(v):
        return '%s(_%s)' % (cxxShortId(v.name), cxxShortId(v.name))
    params  = reduce(lambda x,y: string.join([x,y],', '),
                     map(mkArg, e.variables))
    assigns = reduce(lambda x,y: string.join([x,y],', '),
                     map(mkAssign, e.variables))
    st.out("""\
@id@(@params@)
    : @assigns@ {}
""", id=cxxShortId(e.id), params=params, assigns=assigns)

    # parameters:
    for v in e.variables:
        st.out('@type@ @name@;', type=cxxTypeStorage(v.type, e.id), name=cxxShortId(v.name))

    st.dec_indent()
    st.out('};')
    needsToGeneratePart2.append(e)

def generateStructPart2(st, e):
    st.out("""\
namespace upf {

inline Arguments& operator<<(Arguments& holder, const @type@& value)
{""", type=cxxFullId(e.id))
    for v in e.variables:
        st.out('    holder << value.@name@;', name=cxxShortId(v.name))
    st.out("""\
    return holder;
}

inline Arguments& operator>>(Arguments& holder, @type@& value)
{""", type=cxxFullId(e.id))
    for v in e.variables:
        st.out('    holder >> value.@name@;', name=cxxShortId(v.name))
    st.out("""\
    return holder;
}

} // namespace upf
""")


def __unionValueTrans(e):
    if cxxGetTypeInfo(e.switchType) == TYPE_PRIMITIVE:
       return lambda x: x
    else:
       return lambda x: cxxFullId(x)

def generateUnion(st, e):
    st.out("""\
struct @id@
{
    @id@() {}
    @des@ _d;
    private:
    """, id=cxxShortId(e.id), des=cxxTypeStorage(e.switchType))
    st.inc_indent()

    for c in e.cases:
        st.out('@type@ _@name@;', type=cxxTypeStorage(c.type), name=cxxShortId(c.name))

    st.out('public:')

    # convenience operators:
    for c in e.cases:
        st.out("""\
void @name@(@type@ _v)
{
    _d = @value@;
    _@name@ = _v;
}
@typeRet@ @name@() const
{
    assert(_d == @value@);
    return _@name@;
}
""",
            id=cxxShortId(e.id),
            value=__unionValueTrans(e)(c.value),
            type=cxxTypeArg(c.type, out=0, context=e.id),
            typeRet=cxxTypeRet(c.type),
            name=cxxShortId(c.name))

    st.dec_indent()
    st.out('};')
    needsToGeneratePart2.append(e)

def generateUnionPart2(st, e):    
    assertList = []
    valueTrans = __unionValueTrans(e)
    for c in e.cases:
        assertList.append('value._d == %s' % valueTrans(c.value))
    assertStr = string.join(assertList, ' || ')
    st.out("""\
namespace upf {

inline Arguments& operator<<(Arguments& holder, const @type@& value)
{
    holder << value._d;
    switch (value._d)
    {""", type=cxxFullId(e.id))
    st.inc_indent()
    for c in e.cases:
        st.out("""\
case @value@:
    holder << value.@name@();
    break;""", value=valueTrans(c.value), name=cxxShortId(c.name))
    st.dec_indent()
    st.out("""\
    default:
        assert(@asse@);
    }
    return holder;
}

inline Arguments& operator>>(Arguments& holder, @type@& value)
{
    holder >> value._d;
    switch (value._d)
    {""", type=cxxFullId(e.id), asse=assertStr)
    st.inc_indent()
    for c in e.cases:
        st.out("""\
case @value@:
    {
    @type@ tmp;
    holder >> tmp;
    value.@name@(tmp);
    break;
    }""", value=valueTrans(c.value), name=cxxShortId(c.name),
          type=cxxTypeStorage(c.type))
    st.dec_indent()
    st.out("""\
    default:
        assert(@asse@);
    }
    return holder;
}

} // namespace upf
""", asse=assertStr)



def generateEntity(st, e):
    if isinstance(e, InterfaceEntity):
        generateInterface(st, e)
    elif isinstance(e, TypedefEntity):
        generateTypedef(st, e)
    elif isinstance(e, StructEntity):
        generateStruct(st, e)
    elif isinstance(e, EnumEntity):
        generateEnum(st, e)
    elif isinstance(e, UnionEntity):
        generateUnion(st, e)
    elif isinstance(e, ConstEntity):
        generateConst(st, e)


def generateEntityPart2(st, e):
    if isinstance(e, InterfaceEntity):
        generateInterfacePart2(st, e)
    elif isinstance(e, StructEntity):
        generateStructPart2(st, e)
    elif isinstance(e, UnionEntity):
        generateUnionPart2(st, e)
    elif isinstance(e, EnumEntity):
        generateEnumPart2(st, e)


def generate(file, outfile):
    doc = ParsedFile()
    doc.parseFile(file)
    global allEntities
    allEntities = doc.allEntities
    st = output.Stream(outfile)
    
    st.out("""\
/*
 * Generated by upf-idl C++ backend (upf ABI version: @abiver@)
 * Do not edit by hand!
 */

#ifndef _UPF_@filename@_H_
#define _UPF_@filename@_H_

#include <upf/cxx.h>

#if UPF_ABI_VERSION != @abiver@
#error "Headers generated for upf ABI different from the one you have installed, please regenerate them."
#endif

""", filename=cxxMkMacro(doc.idlFile), abiver=GENERATED_FOR_ABI_VERSION)

    for pragma in doc.pragmas:
        comp = pragma.text.split(' ')
        if len(comp) == 2 and comp[0] == 'cxx_include':
            st.out('#include @file@', file=comp[1])

    for name,local in doc.includes:
        if local:
            st.out('#include "@file@.h"', file=os.path.splitext(name)[0])
        else:
            st.out('#include <@file@.h>', file=os.path.splitext(name)[0])

    for e in doc.toplevel:
        NS = e.id.split('.')[:-1]
        for n in NS:
            st.out('namespace @name@ \n{', name=n)
            st.inc_indent()
        # dump forward references that belong into this namespace:
        for fwd in filter(lambda x: x.split('.')[:-1] == NS,
                          doc.forwardRefs):
            st.out("class @id@;", id=cxxShortId(fwd))
            doc.forwardRefs.remove(fwd)
        # generate the entity:
        generateEntity(st, e)
        for n in NS:
            st.dec_indent()
            st.out('}')

    if len(needsToGeneratePart2) > 0:
        st.out('#include <upf/cxx_ptr.h>')
        for e in needsToGeneratePart2:
            generateEntityPart2(st, e)
    
    st.out('\n#endif')
